@@ -36,6 +36,7 @@
 #include <pin.h>
 #include <pkcs11_int.h>
 #include <p11-kit/p11-kit.h>
+#include "pkcs11x.h"
 #include <p11-kit/pin.h>
 #ifdef HAVE_GETPID
 # include <unistd.h>
@@ -43,8 +44,6 @@
 
 #define MAX_PROVIDERS 16
 
-/* XXX: try to eliminate this */
-#define MAX_CERT_SIZE 32*1024
 #define MAX_SLOTS 48
 
 extern void *_gnutls_pkcs11_mutex;
@@ -69,7 +68,6 @@ struct find_url_data_st {
 
 struct find_obj_data_st {
 	gnutls_pkcs11_obj_t *p_list;
-	unsigned int *n_list;
 	unsigned int current;
 	gnutls_pkcs11_obj_attr_t flags;
 	struct p11_kit_uri *info;
@@ -717,7 +715,7 @@ gnutls_pkcs11_set_token_function(gnutls_pkcs11_token_callback_t fn,
 	_gnutls_token_data = userdata;
 }
 
-int pkcs11_url_to_info(const char *url, struct p11_kit_uri **info)
+int pkcs11_url_to_info(const char *url, struct p11_kit_uri **info, unsigned flags)
 {
 	int allocated = 0;
 	int ret;
@@ -742,6 +740,26 @@ int pkcs11_url_to_info(const char *url, struct p11_kit_uri **info)
 		    GNUTLS_E_MEMORY_ERROR : GNUTLS_E_PARSING_ERROR;
 	}
 
+	/* check for incomplete URIs */
+	if (p11_kit_uri_get_attribute (*info, CKA_CLASS) == NULL) {
+		struct ck_attribute at;
+		ck_object_class_t klass;
+
+		if (flags & GNUTLS_PKCS11_OBJ_FLAG_EXPECT_CERT) {
+			klass = CKO_CERTIFICATE;
+			at.type = CKA_CLASS;
+			at.value = &klass;
+			at.value_len = sizeof (klass);
+			p11_kit_uri_set_attribute (*info, &at);
+		} else if (flags & GNUTLS_PKCS11_OBJ_FLAG_EXPECT_PRIVKEY) {
+			klass = CKO_PRIVATE_KEY;
+			at.type = CKA_CLASS;
+			at.value = &klass;
+			at.value_len = sizeof (klass);
+			p11_kit_uri_set_attribute (*info, &at);
+		}
+	}
+
 	return 0;
 }
 
@@ -1068,13 +1086,14 @@ pkcs11_open_session(struct pkcs11_session_info *sinfo,
 	/* ok found */
 	sinfo->pks = pks;
 	sinfo->module = module;
+	sinfo->sid = slot;
 	sinfo->init = 1;
 	memcpy(&sinfo->tinfo, &tinfo.tinfo, sizeof(sinfo->tinfo));
 
 	if (flags & SESSION_LOGIN) {
 		ret =
-		    pkcs11_login(sinfo, pin_info, &tinfo, info,
-				 (flags & SESSION_SO) ? 1 : 0);
+		    pkcs11_login(sinfo, pin_info, info,
+				 (flags & SESSION_SO) ? 1 : 0, 0);
 		if (ret < 0) {
 			gnutls_assert();
 			pkcs11_close_session(sinfo);
@@ -1150,11 +1169,14 @@ _pkcs11_traverse_tokens(find_func_t find_func, void *input,
 
 			sinfo.module = module;
 			sinfo.pks = pks;
+			sinfo.sid = tinfo.sid;
+			memcpy(&sinfo.tinfo, &tinfo.tinfo, sizeof(sinfo.tinfo));
 
 			if (flags & SESSION_LOGIN) {
 				ret =
-				    pkcs11_login(&sinfo, pin_info, &tinfo,
-						 info, (flags & SESSION_SO) ? 1 :	 0);
+				    pkcs11_login(&sinfo, pin_info,
+						 info, (flags & SESSION_SO) ? 1 : 0,
+						 0);
 				if (ret < 0) {
 					gnutls_assert();
 					return ret;
@@ -1213,6 +1235,9 @@ pkcs11_obj_import(ck_object_class_t class, gnutls_pkcs11_obj_t obj,
 	case CKO_CERTIFICATE:
 		obj->type = GNUTLS_PKCS11_OBJ_X509_CRT;
 		break;
+	case CKO_X_CERTIFICATE_EXTENSION:
+		obj->type = GNUTLS_PKCS11_OBJ_X509_CRT_EXTENSION;
+		break;
 	case CKO_PUBLIC_KEY:
 		obj->type = GNUTLS_PKCS11_OBJ_PUBKEY;
 		break;
@@ -1422,10 +1447,11 @@ cleanup:
 static int
 pkcs11_obj_import_pubkey(struct ck_function_list *module,
 			 ck_session_handle_t pks,
-			 ck_object_handle_t obj,
-			 gnutls_pkcs11_obj_t crt,
-			 const gnutls_datum_t * id,
-			 const gnutls_datum_t * label,
+			 ck_object_handle_t ctx,
+			 gnutls_pkcs11_obj_t pobj,
+			 gnutls_datum_t *data,
+			 const gnutls_datum_t *id,
+			 const gnutls_datum_t *label,
 			 struct ck_token_info *tinfo,
 			 struct ck_info *lib_info)
 {
@@ -1438,12 +1464,12 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &key_type;
 	a[0].value_len = sizeof(key_type);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
-		crt->pk_algorithm = mech_to_pk(key_type);
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
+		pobj->pk_algorithm = mech_to_pk(key_type);
 
 		ret =
-		    pkcs11_read_pubkey(module, pks, obj, key_type,
-				       crt->pubkey);
+		    pkcs11_read_pubkey(module, pks, ctx, key_type,
+				       pobj->pubkey);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 	}
@@ -1453,9 +1479,9 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &tval;
 	a[0].value_len = sizeof(tval);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
 		if (tval != 0) {
-			crt->key_usage |= GNUTLS_KEY_DATA_ENCIPHERMENT;
+			pobj->key_usage |= GNUTLS_KEY_DATA_ENCIPHERMENT;
 		}
 	}
 
@@ -1463,9 +1489,9 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &tval;
 	a[0].value_len = sizeof(tval);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
 		if (tval != 0) {
-			crt->key_usage |= GNUTLS_KEY_DIGITAL_SIGNATURE |
+			pobj->key_usage |= GNUTLS_KEY_DIGITAL_SIGNATURE |
 			    GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN
 			    | GNUTLS_KEY_NON_REPUDIATION;
 		}
@@ -1475,9 +1501,9 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &tval;
 	a[0].value_len = sizeof(tval);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
 		if (tval != 0) {
-			crt->key_usage |= GNUTLS_KEY_DIGITAL_SIGNATURE |
+			pobj->key_usage |= GNUTLS_KEY_DIGITAL_SIGNATURE |
 			    GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN
 			    | GNUTLS_KEY_NON_REPUDIATION;
 		}
@@ -1487,9 +1513,9 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &tval;
 	a[0].value_len = sizeof(tval);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
 		if (tval != 0) {
-			crt->key_usage |= GNUTLS_KEY_KEY_AGREEMENT;
+			pobj->key_usage |= GNUTLS_KEY_KEY_AGREEMENT;
 		}
 	}
 
@@ -1497,21 +1523,22 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 	a[0].value = &tval;
 	a[0].value_len = sizeof(tval);
 
-	if (pkcs11_get_attribute_value(module, pks, obj, a, 1) == CKR_OK) {
+	if (pkcs11_get_attribute_value(module, pks, ctx, a, 1) == CKR_OK) {
 		if (tval != 0) {
-			crt->key_usage |= GNUTLS_KEY_KEY_ENCIPHERMENT;
+			pobj->key_usage |= GNUTLS_KEY_KEY_ENCIPHERMENT;
 		}
 	}
 
-	return pkcs11_obj_import(CKO_PUBLIC_KEY, crt, NULL, id, label,
+	ret = pkcs11_obj_import(CKO_PUBLIC_KEY, pobj, data, id, label,
 				 tinfo, lib_info);
+	return ret;
 }
 
 static int
 pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 		     struct pkcs11_session_info *sinfo,
 		     struct token_info *info, struct ck_info *lib_info,
-		     gnutls_pkcs11_obj_t fobj)
+		     gnutls_pkcs11_obj_t pobj)
 {
 	ck_bool_t b;
 	int rv, ret;
@@ -1519,18 +1546,17 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 	unsigned long category = 0;
 	char label_tmp[PKCS11_LABEL_SIZE];
 	char id_tmp[PKCS11_ID_SIZE];
-	uint8_t *cert_data = NULL;
-	gnutls_datum_t id, label, data;
+	gnutls_datum_t id, label, data = {NULL, 0};
 
 	/* now figure out flags */
-	fobj->flags = 0;
+	pobj->flags = 0;
 	a[0].type = CKA_WRAP;
 	a[0].value = &b;
 	a[0].value_len = sizeof(b);
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && b != 0)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_KEY_WRAP;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_KEY_WRAP;
 
 	a[0].type = CKA_UNWRAP;
 	a[0].value = &b;
@@ -1538,7 +1564,7 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && b != 0)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_KEY_WRAP;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_KEY_WRAP;
 
 	a[0].type = CKA_PRIVATE;
 	a[0].value = &b;
@@ -1546,7 +1572,7 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && b != 0)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_PRIVATE;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_PRIVATE;
 
 	a[0].type = CKA_TRUSTED;
 	a[0].value = &b;
@@ -1554,7 +1580,7 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && b != 0)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_TRUSTED;
 
 	a[0].type = CKA_SENSITIVE;
 	a[0].value = &b;
@@ -1562,7 +1588,7 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && b != 0)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE;
 
 	a[0].type = CKA_CERTIFICATE_CATEGORY;
 	a[0].value = &category;
@@ -1570,46 +1596,45 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	rv = pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv == CKR_OK && category == 2)
-    		fobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_CA;
+    		pobj->flags |= GNUTLS_PKCS11_OBJ_FLAG_MARK_CA;
 
 	/* now recover the object label/id */
 	a[0].type = CKA_LABEL;
 	a[0].value = label_tmp;
 	a[0].value_len = sizeof(label_tmp);
-	a[1].type = CKA_ID;
-	a[1].value = id_tmp;
-	a[1].value_len = sizeof(id_tmp);
-
 	rv = pkcs11_get_attribute_value
-	    (sinfo->module, sinfo->pks, obj, a, 2);
+	    (sinfo->module, sinfo->pks, obj, a, 1);
 	if (rv != CKR_OK) {
 		gnutls_assert();
-		return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
-	}
-
-	id.data = a[1].value;
-	id.size = a[1].value_len;
-	label.data = a[0].value;
-	label.size = a[0].value_len;
-
-	cert_data = gnutls_malloc(MAX_CERT_SIZE);
-	if (cert_data == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_MEMORY_ERROR;
-	}
-
-	a[0].type = CKA_VALUE;
-	a[0].value = cert_data;
-	a[0].value_len = MAX_CERT_SIZE;
-
-	rv = pkcs11_get_attribute_value
-	    (sinfo->module, sinfo->pks, obj, a, 2);
-	if (rv == CKR_OK) {
-		data.data = a[0].value;
-		data.size = a[0].value_len;
+		label.data = NULL;
+		label.size = 0;
 	} else {
-		data.data = NULL;
-		data.size = 0;
+		label.data = a[0].value;
+		label.size = a[0].value_len;
+	}
+
+	a[0].type = CKA_ID;
+	a[0].value = id_tmp;
+	a[0].value_len = sizeof(id_tmp);
+	rv = pkcs11_get_attribute_value
+	    (sinfo->module, sinfo->pks, obj, a, 1);
+	if (rv != CKR_OK) {
+		gnutls_assert();
+		id.data = NULL;
+		id.size = 0;
+	} else {
+		id.data = a[0].value;
+		id.size = a[0].value_len;
+	}
+
+	if (label.data == NULL && id.data == NULL)
+		return gnutls_assert_val(GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
+
+	rv = pkcs11_get_attribute_avalue
+	    (sinfo->module, sinfo->pks, obj, CKA_VALUE, &data);
+	if (rv != CKR_OK) {
+		gnutls_assert();
+		/* data will be null */
 	}
 
 	if (class == CKO_PUBLIC_KEY) {
@@ -1617,14 +1642,15 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 		    pkcs11_obj_import_pubkey(sinfo->module,
 					     sinfo->pks,
 					     obj,
-					     fobj,
+					     pobj,
+					     &data,
 					     &id, &label,
 					     &info->tinfo,
 					     lib_info);
 	} else {
 		ret =
 		    pkcs11_obj_import(class,
-				      fobj,
+				      pobj,
 				      &data, &id, &label,
 				      &info->tinfo,
 				      lib_info);
@@ -1636,12 +1662,12 @@ pkcs11_import_object(ck_object_handle_t obj, ck_object_class_t class,
 
 	ret = 0;
  cleanup:
- 	gnutls_free(cert_data);
+ 	gnutls_free(data.data);
  	return ret;
 }
 
 static int
-find_obj_url(struct pkcs11_session_info *sinfo,
+find_obj_url_cb(struct pkcs11_session_info *sinfo,
 	     struct token_info *info, struct ck_info *lib_info,
 	     void *input)
 {
@@ -1719,7 +1745,6 @@ find_obj_url(struct pkcs11_session_info *sinfo,
 
 	if (pkcs11_find_objects(sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK &&
 	    count == 1) {
-
 		ret = pkcs11_import_object(obj, class, sinfo, info, lib_info, find_data->obj);
 		if (ret >= 0) {
 			found = 1;
@@ -1785,14 +1810,14 @@ gnutls_pkcs11_obj_import_url(gnutls_pkcs11_obj_t obj, const char *url,
 	/* fill in the find data structure */
 	find_data.obj = obj;
 
-	ret = pkcs11_url_to_info(url, &obj->info);
+	ret = pkcs11_url_to_info(url, &obj->info, flags);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
 	}
 
 	ret =
-	    _pkcs11_traverse_tokens(find_obj_url, &find_data, obj->info,
+	    _pkcs11_traverse_tokens(find_obj_url_cb, &find_data, obj->info,
 				    &obj->pin,
 				    pkcs11_obj_flags_to_int(flags));
 	if (ret < 0) {
@@ -1804,7 +1829,7 @@ gnutls_pkcs11_obj_import_url(gnutls_pkcs11_obj_t obj, const char *url,
 }
 
 static int
-find_token_num(struct pkcs11_session_info *sinfo,
+find_token_num_cb(struct pkcs11_session_info *sinfo,
 	       struct token_info *tinfo,
 	       struct ck_info *lib_info, void *input)
 {
@@ -1858,7 +1883,7 @@ gnutls_pkcs11_token_get_url(unsigned int seq,
 	tn.seq = seq;
 	tn.info = p11_kit_uri_new();
 
-	ret = _pkcs11_traverse_tokens(find_token_num, &tn, NULL, NULL, 0);
+	ret = _pkcs11_traverse_tokens(find_token_num_cb, &tn, NULL, NULL, 0);
 	if (ret < 0) {
 		p11_kit_uri_free(tn.info);
 		gnutls_assert();
@@ -1905,7 +1930,7 @@ gnutls_pkcs11_token_get_info(const char *url,
 
 	PKCS11_CHECK_INIT;
 
-	ret = pkcs11_url_to_info(url, &info);
+	ret = pkcs11_url_to_info(url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
@@ -2096,8 +2121,11 @@ retrieve_pin_from_callback(const struct pin_info_st *pin_info,
 		return GNUTLS_E_MEMORY_ERROR;
 	}
 
-	if (user_type == CKU_USER) {
+	if (user_type == CKU_USER || user_type == CKU_CONTEXT_SPECIFIC) {
 		flags |= GNUTLS_PIN_USER;
+
+		if (user_type == CKU_CONTEXT_SPECIFIC)
+			flags |= GNUTLS_PIN_CONTEXT_SPECIFIC;
 		if (token_info->flags & CKF_USER_PIN_COUNT_LOW)
 			flags |= GNUTLS_PIN_COUNT_LOW;
 		if (token_info->flags & CKF_USER_PIN_FINAL_TRY)
@@ -2180,16 +2208,24 @@ retrieve_pin(struct pin_info_st *pin_info, struct p11_kit_uri *info,
 int
 pkcs11_login(struct pkcs11_session_info *sinfo,
 	     struct pin_info_st *pin_info,
-	     const struct token_info *tokinfo, struct p11_kit_uri *info,
-	     int so)
+	     struct p11_kit_uri *info,
+	     unsigned so,
+	     unsigned reauth)
 {
 	struct ck_session_info session_info;
 	int attempt = 0, ret;
 	ck_user_type_t user_type;
 	ck_rv_t rv;
 
-	user_type = (so == 0) ? CKU_USER : CKU_SO;
-	if (so == 0 && (tokinfo->tinfo.flags & CKF_LOGIN_REQUIRED) == 0) {
+	if (so == 0) {
+		if (reauth == 0)
+			user_type = CKU_USER;
+		else
+			user_type = CKU_CONTEXT_SPECIFIC;
+	} else
+		user_type = CKU_SO;
+
+	if (so == 0 && (sinfo->tinfo.flags & CKF_LOGIN_REQUIRED) == 0) {
 		gnutls_assert();
 		_gnutls_debug_log("p11: No login required.\n");
 		return 0;
@@ -2198,10 +2234,9 @@ pkcs11_login(struct pkcs11_session_info *sinfo,
 	/* For a token with a "protected" (out-of-band) authentication
 	 * path, calling login with a NULL username is all that is
 	 * required. */
-	if (tokinfo->tinfo.flags & CKF_PROTECTED_AUTHENTICATION_PATH) {
+	if (sinfo->tinfo.flags & CKF_PROTECTED_AUTHENTICATION_PATH) {
 		rv = (sinfo->module)->C_Login(sinfo->pks,
-					      (so ==
-					       0) ? CKU_USER : CKU_SO,
+					      user_type,
 					      NULL, 0);
 		if (rv == CKR_OK || rv == CKR_USER_ALREADY_LOGGED_IN) {
 			return 0;
@@ -2218,13 +2253,13 @@ pkcs11_login(struct pkcs11_session_info *sinfo,
 		struct p11_kit_pin *pin;
 		struct ck_token_info tinfo;
 
-		memcpy(&tinfo, &tokinfo->tinfo, sizeof(tinfo));
+		memcpy(&tinfo, &sinfo->tinfo, sizeof(tinfo));
 
 		/* Check whether the session is already logged in, and if so, just skip */
 		rv = (sinfo->module)->C_GetSessionInfo(sinfo->pks,
 						       &session_info);
-		if (rv == CKR_OK
-		    && (session_info.state == CKS_RO_USER_FUNCTIONS
+		if (rv == CKR_OK && reauth == 0 &&
+		    (session_info.state == CKS_RO_USER_FUNCTIONS
 			|| session_info.state == CKS_RW_USER_FUNCTIONS)) {
 			ret = 0;
 			goto cleanup;
@@ -2234,7 +2269,7 @@ pkcs11_login(struct pkcs11_session_info *sinfo,
 		 * status again, the flags might change. */
 		if (attempt) {
 			if (pkcs11_get_token_info
-			    (tokinfo->prov->module, tokinfo->sid,
+			    (sinfo->module, sinfo->sid,
 			     &tinfo) != CKR_OK) {
 				gnutls_assert();
 				_gnutls_debug_log
@@ -2378,9 +2413,10 @@ find_privkeys(struct pkcs11_session_info *sinfo,
 
 /* Recover certificate list from tokens */
 
+#define OBJECTS_A_TIME 8*1024
 
 static int
-find_objs(struct pkcs11_session_info *sinfo,
+find_objs_cb(struct pkcs11_session_info *sinfo,
 	  struct token_info *info, struct ck_info *lib_info, void *input)
 {
 	struct find_obj_data_st *find_data = input;
@@ -2391,22 +2427,16 @@ find_objs(struct pkcs11_session_info *sinfo,
 	ck_bool_t trusted;
 	unsigned long category;
 	ck_rv_t rv;
-	ck_object_handle_t obj;
+	ck_object_handle_t *objs = NULL;
 	unsigned long count;
 	char certid_tmp[PKCS11_ID_SIZE];
 	int ret;
 	struct find_pkey_list_st plist;	/* private key holder */
 	unsigned int i, tot_values = 0;
 
-	if (info == NULL) {	/* final call */
-		if (find_data->current <= *find_data->n_list)
-			ret = 0;
-		else
-			ret = GNUTLS_E_SHORT_MEMORY_BUFFER;
-
-		*find_data->n_list = find_data->current;
-
-		return ret;
+	if (info == NULL) {
+		gnutls_assert();
+		return 0;
 	}
 
 	/* do not bother reading the token if basic fields do not match
@@ -2471,21 +2501,6 @@ find_objs(struct pkcs11_session_info *sinfo,
 			tot_values++;
 		}
 
-		attr = p11_kit_uri_get_attribute(find_data->info, CKA_ID);
-		if (attr) {
-			a[tot_values].type = CKA_ID;
-			a[tot_values].value = attr->value;
-			a[tot_values].value_len = attr->value_len;
-			tot_values++;
-		}
-
-		attr = p11_kit_uri_get_attribute(find_data->info, CKA_LABEL);
-		if (attr) {
-			a[tot_values].type = CKA_LABEL;
-			a[tot_values].value = attr->value;
-			a[tot_values].value_len = attr->value_len;
-			tot_values++;
-		}
 	} else if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED) {
 		class = CKO_CERTIFICATE;
 		type = CKC_X_509;
@@ -2557,7 +2572,17 @@ find_objs(struct pkcs11_session_info *sinfo,
 
 	attr = p11_kit_uri_get_attribute(find_data->info, CKA_ID);
 	if (attr != NULL) {
-		memcpy(a + tot_values, attr, sizeof(struct ck_attribute));
+		a[tot_values].type = CKA_ID;
+		a[tot_values].value = attr->value;
+		a[tot_values].value_len = attr->value_len;
+		tot_values++;
+	}
+
+	attr = p11_kit_uri_get_attribute(find_data->info, CKA_LABEL);
+	if (attr) {
+		a[tot_values].type = CKA_LABEL;
+		a[tot_values].value = attr->value;
+		a[tot_values].value_len = attr->value_len;
 		tot_values++;
 	}
 
@@ -2569,53 +2594,64 @@ find_objs(struct pkcs11_session_info *sinfo,
 		return pkcs11_rv_to_err(rv);
 	}
 
+	objs = gnutls_malloc(OBJECTS_A_TIME*sizeof(objs[0]));
+	if (objs == NULL) {
+		ret = gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
+		goto fail;
+	}
+
 	while (pkcs11_find_objects
-	       (sinfo->module, sinfo->pks, &obj, 1, &count) == CKR_OK
-	       && count == 1) {
+	       (sinfo->module, sinfo->pks, objs, OBJECTS_A_TIME, &count) == CKR_OK
+	       && count > 0) {
+	        unsigned j;
 	        gnutls_datum_t id;
 
-		a[0].type = CKA_ID;
-		a[0].value = certid_tmp;
-		a[0].value_len = sizeof certid_tmp;
-
-		if (pkcs11_get_attribute_value
-		    (sinfo->module, sinfo->pks, obj, a, 1) == CKR_OK) {
-			id.data = a[0].value;
-			id.size = a[0].value_len;
-		} else {
-			id.data = NULL;
-			id.size = 0;
+		find_data->p_list = gnutls_realloc_fast(find_data->p_list, (find_data->current+count)*sizeof(find_data->p_list[0]));
+		if (find_data->p_list == NULL) {
+			ret = gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
+			goto fail;
 		}
-		tot_values++;
 
-		if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_ALL ||
-			find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_MATCH) {
-			a[0].type = CKA_CLASS;
-			a[0].value = &class;
-			a[0].value_len = sizeof class;
+	        for (j=0;j<count;j++) {
+			a[0].type = CKA_ID;
+			a[0].value = certid_tmp;
+			a[0].value_len = sizeof certid_tmp;
 
-			rv = pkcs11_get_attribute_value(sinfo->module,
-						   sinfo->pks, obj, a, 1);
-			if (rv != CKR_OK) {
-				class = -1;
+			if (pkcs11_get_attribute_value
+			    (sinfo->module, sinfo->pks, objs[j], a, 1) == CKR_OK) {
+				id.data = a[0].value;
+				id.size = a[0].value_len;
+			} else {
+				id.data = NULL;
+				id.size = 0;
 			}
-		}
 
-		if (find_data->flags ==
-		    GNUTLS_PKCS11_OBJ_ATTR_CRT_WITH_PRIVKEY) {
-			for (i = 0; i < plist.key_ids_size; i++) {
-				if (plist.key_ids[i].length !=
-				    id.size
-				    || memcmp(plist.key_ids[i].data,
-					      id.data,
-					      id.size) != 0) {
-					/* not found */
-					continue;
+			if (find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_ALL ||
+			    find_data->flags == GNUTLS_PKCS11_OBJ_ATTR_MATCH) {
+				a[0].type = CKA_CLASS;
+				a[0].value = &class;
+				a[0].value_len = sizeof class;
+
+				rv = pkcs11_get_attribute_value(sinfo->module,
+							   sinfo->pks, objs[j], a, 1);
+				if (rv != CKR_OK) {
+					class = -1;
 				}
 			}
-		}
 
-		if (find_data->current < *find_data->n_list) {
+			if (find_data->flags ==
+			    GNUTLS_PKCS11_OBJ_ATTR_CRT_WITH_PRIVKEY) {
+				for (i = 0; i < plist.key_ids_size; i++) {
+					if (plist.key_ids[i].length !=
+					    id.size
+					    || memcmp(plist.key_ids[i].data,
+						      id.data,
+						      id.size) != 0) {
+						/* not found */
+						continue;
+					}
+ 				}
+ 			}
 
 			ret =
 			    gnutls_pkcs11_obj_init(&find_data->p_list
@@ -2625,24 +2661,26 @@ find_objs(struct pkcs11_session_info *sinfo,
 				goto fail;
 			}
 
-			ret = pkcs11_import_object(obj, class, sinfo,
+			ret = pkcs11_import_object(objs[j], class, sinfo,
 					     info, lib_info,
 					     find_data->p_list[find_data->current]);
 			if (ret < 0) {
 				gnutls_assert();
-				goto fail;
+				/* skip the failed object */
+				continue;
 			}
 
-		}
-
-		find_data->current++;
+			find_data->current++;
+ 		}
 	}
 
+	gnutls_free(objs);
 	pkcs11_find_objects_final(sinfo);
 
 	return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;	/* continue until all tokens have been checked */
 
       fail:
+	gnutls_free(objs);
 	pkcs11_find_objects_final(sinfo);
 	if (plist.key_ids != NULL) {
 		for (i = 0; i < plist.key_ids_size; i++) {
@@ -2650,9 +2688,13 @@ find_objs(struct pkcs11_session_info *sinfo,
 		}
 		gnutls_free(plist.key_ids);
 	}
-	for (i = 0; i < find_data->current; i++) {
-		gnutls_pkcs11_obj_deinit(find_data->p_list[i]);
+	if (find_data->p_list != NULL) {
+		for (i = 0; i < find_data->current; i++) {
+			gnutls_pkcs11_obj_deinit(find_data->p_list[i]);
+		}
+		gnutls_free(find_data->p_list);
 	}
+	find_data->p_list = NULL;
 	find_data->current = 0;
 
 	return ret;
@@ -2683,29 +2725,27 @@ gnutls_pkcs11_obj_list_import_url(gnutls_pkcs11_obj_t * p_list,
 {
 	int ret;
 	struct find_obj_data_st priv;
+	unsigned i;
 
 	PKCS11_CHECK_INIT;
 
 	memset(&priv, 0, sizeof(priv));
 
 	/* fill in the find data structure */
-	priv.p_list = p_list;
-	priv.n_list = n_list;
 	priv.flags = attrs;
-	priv.current = 0;
 
 	if (url == NULL || url[0] == 0) {
 		url = "pkcs11:";
 	}
 
-	ret = pkcs11_url_to_info(url, &priv.info);
+	ret = pkcs11_url_to_info(url, &priv.info, flags);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
 	}
 
 	ret =
-	    _pkcs11_traverse_tokens(find_objs, &priv, priv.info,
+	    _pkcs11_traverse_tokens(find_objs_cb, &priv, priv.info,
 				    NULL, pkcs11_obj_flags_to_int(flags));
 	p11_kit_uri_free(priv.info);
 
@@ -2718,6 +2758,19 @@ gnutls_pkcs11_obj_list_import_url(gnutls_pkcs11_obj_t * p_list,
 		return ret;
 	}
 
+	if (priv.current > *n_list) {
+		*n_list = priv.current;
+		for (i=0;i<priv.current;i++) {
+			gnutls_pkcs11_obj_deinit(priv.p_list[i]);
+		}
+		gnutls_free(priv.p_list);
+		return gnutls_assert_val(GNUTLS_E_SHORT_MEMORY_BUFFER);
+	}
+
+	*n_list = priv.current;
+	memcpy(p_list, priv.p_list, priv.current*sizeof(p_list[0]));
+	gnutls_free(priv.p_list);
+
 	return 0;
 }
 
@@ -2733,6 +2786,9 @@ gnutls_pkcs11_obj_list_import_url(gnutls_pkcs11_obj_t * p_list,
  * by using all objects identified by the PKCS 11 URL. The output
  * is stored in @p_list, which will be initialized.
  *
+ * All returned objects must be deinitialized using gnutls_pkcs11_obj_deinit(),
+ * and @p_list must be free'd using gnutls_free().
+ *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
@@ -2745,41 +2801,44 @@ gnutls_pkcs11_obj_list_import_url2(gnutls_pkcs11_obj_t ** p_list,
 				   gnutls_pkcs11_obj_attr_t attrs,
 				   unsigned int flags)
 {
-	unsigned int init = 128;
 	int ret;
+	struct find_obj_data_st priv;
 
-	*p_list = gnutls_malloc(sizeof(gnutls_pkcs11_obj_t) * init);
-	if (*p_list == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_MEMORY_ERROR;
-	}
+	PKCS11_CHECK_INIT;
 
-	ret =
-	    gnutls_pkcs11_obj_list_import_url(*p_list, &init, url, attrs,
-					      flags);
-	if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER) {
-		*p_list =
-		    gnutls_realloc_fast(*p_list,
-					sizeof(gnutls_pkcs11_obj_t) *
-					init);
-		if (*p_list == NULL)
-			return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
+	memset(&priv, 0, sizeof(priv));
 
-		ret =
-		    gnutls_pkcs11_obj_list_import_url(*p_list, &init, url,
-						      attrs, flags);
+	/* fill in the find data structure */
+	priv.flags = attrs;
+
+	if (url == NULL || url[0] == 0) {
+		url = "pkcs11:";
 	}
 
+	ret = pkcs11_url_to_info(url, &priv.info, flags);
 	if (ret < 0) {
 		gnutls_assert();
-		gnutls_free(*p_list);
-		*p_list = NULL;
 		return ret;
 	}
 
-	*n_list = init;
-	return 0;
+	ret =
+	    _pkcs11_traverse_tokens(find_objs_cb, &priv, priv.info,
+				    NULL, pkcs11_obj_flags_to_int(flags));
+	p11_kit_uri_free(priv.info);
 
+	if (ret < 0) {
+		gnutls_assert();
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+			*n_list = 0;
+			ret = 0;
+		}
+		return ret;
+	}
+
+	*n_list = priv.current;
+	*p_list = priv.p_list;
+
+	return 0;
 }
 
 /**
@@ -2814,7 +2873,7 @@ gnutls_x509_crt_import_pkcs11_url(gnutls_x509_crt_t crt,
 		gnutls_pkcs11_obj_set_pin_function(pcrt, crt->pin.cb,
 						   crt->pin.data);
 
-	ret = gnutls_pkcs11_obj_import_url(pcrt, url, flags);
+	ret = gnutls_pkcs11_obj_import_url(pcrt, url, flags|GNUTLS_PKCS11_OBJ_FLAG_EXPECT_CERT);
 	if (ret < 0) {
 		gnutls_assert();
 		goto cleanup;
@@ -2904,7 +2963,7 @@ gnutls_x509_crt_list_import_pkcs11(gnutls_x509_crt_t * certs,
 }
 
 static int
-find_flags(struct pkcs11_session_info *sinfo,
+find_flags_cb(struct pkcs11_session_info *sinfo,
 	   struct token_info *info, struct ck_info *lib_info, void *input)
 {
 	struct find_flags_data_st *find_data = input;
@@ -2953,14 +3012,14 @@ int gnutls_pkcs11_token_get_flags(const char *url, unsigned int *flags)
 	PKCS11_CHECK_INIT;
 
 	memset(&find_data, 0, sizeof(find_data));
-	ret = pkcs11_url_to_info(url, &find_data.info);
+	ret = pkcs11_url_to_info(url, &find_data.info, 0);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
 	}
 
 	ret =
-	    _pkcs11_traverse_tokens(find_flags, &find_data, find_data.info,
+	    _pkcs11_traverse_tokens(find_flags_cb, &find_data, find_data.info,
 				    NULL, 0);
 	p11_kit_uri_free(find_data.info);
 
@@ -3009,7 +3068,7 @@ gnutls_pkcs11_token_get_mechanism(const char *url, unsigned int idx,
 
 	PKCS11_CHECK_INIT;
 
-	ret = pkcs11_url_to_info(url, &info);
+	ret = pkcs11_url_to_info(url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
@@ -3067,19 +3126,76 @@ const char *gnutls_pkcs11_type_get_name(gnutls_pkcs11_obj_type_t type)
 		return "Secret key";
 	case GNUTLS_PKCS11_OBJ_DATA:
 		return "Data";
+	case GNUTLS_PKCS11_OBJ_X509_CRT_EXTENSION:
+		return "X.509 certificate extension";
 	case GNUTLS_PKCS11_OBJ_UNKNOWN:
 	default:
 		return "Unknown";
 	}
 }
 
-#ifndef CKA_X_DISTRUSTED
-# define CKA_X_VENDOR (CKA_VENDOR_DEFINED | 0x58444700UL)
-# define CKA_X_DISTRUSTED (CKA_X_VENDOR + 100)
-#endif
+static
+int check_found_cert(struct find_cert_st *priv, gnutls_datum_t *data, time_t now)
+{
+	gnutls_x509_crt_t tcrt = NULL;
+	int ret;
+
+	ret = gnutls_x509_crt_init(&tcrt);
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	ret = gnutls_x509_crt_import(tcrt, data, GNUTLS_X509_FMT_DER);
+	if (ret < 0) {
+		gnutls_assert();
+		goto cleanup;
+	}
+
+	if (priv->key_id.size > 0 &&
+	    !_gnutls_check_valid_key_id(&priv->key_id, tcrt, now)) {
+		gnutls_assert();
+		ret = -1;
+		goto cleanup;
+	}
+
+	if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE) {
+		if (priv->crt == NULL) {
+			gnutls_assert();
+			ret = -1;
+			goto cleanup;
+		}
+
+		if (_gnutls_check_if_same_cert(priv->crt, tcrt) == 0) {
+			/* doesn't match */
+			ret = -1;
+			goto cleanup;
+		}
+	}
+
+	if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY) {
+		if (priv->crt == NULL) {
+			gnutls_assert();
+			ret = -1;
+			goto cleanup;
+		}
+
+		if (_gnutls_check_if_same_key(priv->crt, tcrt, 1) == 0) {
+			/* doesn't match */
+			ret = -1;
+			goto cleanup;
+		}
+	}
+
+	ret = 0;
+cleanup:
+	if (tcrt != NULL)
+		gnutls_x509_crt_deinit(tcrt);
+	return ret;
+}
 
 static int
-find_cert(struct pkcs11_session_info *sinfo,
+find_cert_cb(struct pkcs11_session_info *sinfo,
 	    struct token_info *info, struct ck_info *lib_info, void *input)
 {
 	struct ck_attribute a[10];
@@ -3089,12 +3205,13 @@ find_cert(struct pkcs11_session_info *sinfo,
 	ck_object_handle_t obj;
 	unsigned long count, a_vals;
 	int found = 0, ret;
-	uint8_t *cert_data = NULL;
 	struct find_cert_st *priv = input;
 	char label_tmp[PKCS11_LABEL_SIZE];
 	char id_tmp[PKCS11_ID_SIZE];
+	gnutls_datum_t data = {NULL, 0};
 	unsigned tries, i, finalized;
 	ck_bool_t trusted = 1;
+	time_t now;
 
 	if (info == NULL) {
 		gnutls_assert();
@@ -3113,14 +3230,6 @@ find_cert(struct pkcs11_session_info *sinfo,
 		priv->serial.size == 0)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
-	/* search the token for the key ID */
-
-	cert_data = gnutls_malloc(MAX_CERT_SIZE);
-	if (cert_data == NULL) {
-		gnutls_assert();
-		return GNUTLS_E_MEMORY_ERROR;
-	}
-
 	/* Find objects with given class and type */
 
 	if (priv->key_id.size > 0 && priv->dn.size > 0)
@@ -3128,6 +3237,7 @@ find_cert(struct pkcs11_session_info *sinfo,
 	else
 		tries = 1;
 
+	now = gnutls_time(0);
 	for (i = 0; i < tries; i++) {
 
 		a_vals = 0;
@@ -3202,33 +3312,50 @@ find_cert(struct pkcs11_session_info *sinfo,
 		       (sinfo->module, sinfo->pks, &obj, 1,
 			&count) == CKR_OK && count == 1) {
 
-			if (priv->need_import == 0 && !(priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE)) {
+			if (priv->need_import == 0 && !(priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE)
+			    && !(priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY)) {
 				found = 1;
 				break;
 			}
 
-			a[0].type = CKA_VALUE;
-			a[0].value = cert_data;
-			a[0].value_len = MAX_CERT_SIZE;
+			a[0].type = CKA_LABEL;
+			a[0].value = label_tmp;
+			a[0].value_len = sizeof(label_tmp);
 
-			a[1].type = CKA_LABEL;
-			a[1].value = label_tmp;
-			a[1].value_len = sizeof(label_tmp);
+			a[1].type = CKA_ID;
+			a[1].value = id_tmp;
+			a[1].value_len = sizeof(id_tmp);
 
-			a[2].type = CKA_ID;
-			a[2].value = id_tmp;
-			a[2].value_len = sizeof(id_tmp);
+			/* data will contain the certificate */
+			rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_VALUE, &data);
 
-			if (pkcs11_get_attribute_value
+			if (rv == CKR_OK && pkcs11_get_attribute_value
 			    (sinfo->module, sinfo->pks, obj, a,
-			     3) == CKR_OK) {
-				gnutls_datum_t id =
-				    { a[2].value, a[2].value_len };
-				gnutls_datum_t data =
-				    { a[0].value, a[0].value_len };
+			     2) == CKR_OK) {
 				gnutls_datum_t label =
+				    { a[0].value, a[0].value_len };
+				gnutls_datum_t id =
 				    { a[1].value, a[1].value_len };
 
+				ret = check_found_cert(priv, &data, now);
+				if (ret < 0) {
+					_gnutls_free_datum(&data);
+					continue;
+				}
+
+				if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_OVERWRITE_TRUSTMOD_EXT) {
+					gnutls_datum_t spki;
+					rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_PUBLIC_KEY_INFO, &spki);
+					if (rv == CKR_OK) {
+						ret = pkcs11_override_cert_exts(sinfo, &spki, &data);
+						gnutls_free(spki.data);
+						if (ret < 0) {
+							gnutls_assert();
+							goto cleanup;
+						}
+					}
+				}
+
 				if (priv->need_import != 0) {
 					ret =
 					    pkcs11_obj_import(class, priv->obj,
@@ -3241,17 +3368,6 @@ find_cert(struct pkcs11_session_info *sinfo,
 					}
 				}
 
-				if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE) {
-					if (priv->crt == NULL) {
-						gnutls_assert();
-						break;
-					}
-
-					if (_gnutls_check_if_same_cert2(priv->crt, &data) == 0) {
-						/* doesn't match */
-						break;
-					}
-				}
 
 				found = 1;
 				break;
@@ -3276,7 +3392,7 @@ find_cert(struct pkcs11_session_info *sinfo,
 	}
 
       cleanup:
-	gnutls_free(cert_data);
+	gnutls_free(data.data);
 	if (finalized == 0)
 		pkcs11_find_objects_final(sinfo);
 
@@ -3287,7 +3403,7 @@ find_cert(struct pkcs11_session_info *sinfo,
  * gnutls_pkcs11_get_raw_issuer:
  * @url: A PKCS 11 url identifying a token
  * @cert: is the certificate to find issuer for
- * @issuer: Will hold the issuer if any in an allocated buffer. 
+ * @issuer: Will hold the issuer if any in an allocated buffer.
  * @fmt: The format of the exported issuer.
  * @flags: Use zero or flags from %GNUTLS_PKCS11_OBJ_FLAG.
  *
@@ -3320,7 +3436,7 @@ int gnutls_pkcs11_get_raw_issuer(const char *url, gnutls_x509_crt_t cert,
 		url = "pkcs11:";
 	}
 
-	ret = pkcs11_url_to_info(url, &info);
+	ret = pkcs11_url_to_info(url, &info, flags);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
@@ -3350,7 +3466,7 @@ int gnutls_pkcs11_get_raw_issuer(const char *url, gnutls_x509_crt_t cert,
 	priv.need_import = 1;
 
 	ret =
-	    _pkcs11_traverse_tokens(find_cert, &priv, info,
+	    _pkcs11_traverse_tokens(find_cert_cb, &priv, info,
 				    NULL, pkcs11_obj_flags_to_int(flags));
 	if (ret < 0) {
 		gnutls_assert();
@@ -3378,13 +3494,13 @@ int gnutls_pkcs11_get_raw_issuer(const char *url, gnutls_x509_crt_t cert,
  * gnutls_pkcs11_crt_is_known:
  * @url: A PKCS 11 url identifying a token
  * @cert: is the certificate to find issuer for
- * @issuer: Will hold the issuer if any in an allocated buffer. 
+ * @issuer: Will hold the issuer if any in an allocated buffer.
  * @fmt: The format of the exported issuer.
  * @flags: Use zero or flags from %GNUTLS_PKCS11_OBJ_FLAG.
  *
  * This function will check whether the provided certificate is stored
  * in the specified token. This is useful in combination with 
- * %GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED or 
+ * %GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED or
  * %GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_DISTRUSTED,
  * to check whether a CA is present or a certificate is blacklisted in
  * a trust PKCS #11 module.
@@ -3419,7 +3535,7 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 		url = "pkcs11:";
 	}
 
-	ret = pkcs11_url_to_info(url, &info);
+	ret = pkcs11_url_to_info(url, &info, 0);
 	if (ret < 0) {
 		gnutls_assert();
 		return 0;
@@ -3431,7 +3547,8 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 	    gnutls_x509_crt_get_serial(cert, serial+sizeof(tag), &serial_size);
 	if (ret < 0) {
 		gnutls_assert();
-		return 0;
+		ret = 0;
+		goto cleanup;
 	}
 
 	/* PKCS#11 requires a DER encoded serial, wtf. $@(*$@ */
@@ -3440,7 +3557,8 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 		tag, &tag_size);
 	if (ret != ASN1_SUCCESS) {
 		gnutls_assert();
-		return 0;
+		ret = 0;
+		goto cleanup;
 	}
 
 	memcpy(serial+sizeof(tag)-tag_size, tag, tag_size);
@@ -3454,16 +3572,17 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 
 	/* when looking for a trusted certificate, we always fully compare
 	 * with the given */
-	if (flags & GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED)
+	if (flags & GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED && !(flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY))
 		flags |= GNUTLS_PKCS11_OBJ_FLAG_COMPARE;
 
 	priv.flags = flags;
 
 	ret =
-	    _pkcs11_traverse_tokens(find_cert, &priv, info,
+	    _pkcs11_traverse_tokens(find_cert_cb, &priv, info,
 				    NULL, pkcs11_obj_flags_to_int(flags));
 	if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
 		/* attempt searching with the subject DN only */
+		gnutls_assert();
 		memset(&priv, 0, sizeof(priv));
 		priv.crt = cert;
 		priv.flags = flags;
@@ -3471,7 +3590,7 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 		priv.dn.data = cert->raw_dn.data;
 		priv.dn.size = cert->raw_dn.size;
 		ret =
-		    _pkcs11_traverse_tokens(find_cert, &priv, info,
+		    _pkcs11_traverse_tokens(find_cert_cb, &priv, info,
 				    NULL, pkcs11_obj_flags_to_int(flags));
 	}
 	if (ret < 0) {
